import * as falafel from 'falafel';
import { ExpressionType } from "./ExpressionType";

export abstract class Expression {
    private static readonly pathExtractor = new RegExp('return (.*);');

    public static path<T>(name: (t: T) => any) {
        const match: RegExpExecArray = Expression.pathExtractor.exec(name + '');
        if (match == null) {
            throw new Error('The function does not contain a statement matching');
        }
        return match[1].split('.').splice(1).join('.');
    }

    public static operator<T>(name: (t: T) => any) {

        const match: RegExpExecArray = Expression.pathExtractor.exec(name + '');

        if (match == null) {
            throw new Error('The function does not contain a statement matching');
        }

        let expr: string = match[1];
        let expType: ExpressionType;


        falafel(expr, function (node: any) {
            if (node.type === 'BinaryExpression' && node.operator == "==") {
                expType = ExpressionType.Equals;
            }

            if (node.type === 'BinaryExpression' && node.operator == "!=") {
                expType = ExpressionType.NotEquals;
            }

            if (node.type === 'BinaryExpression' && node.operator == "<") {
                expType = ExpressionType.LessThan;
            }

            if (node.type === 'BinaryExpression' && node.operator == "<=") {
                expType = ExpressionType.LessThanOrEqual;
            }

            if (node.type === 'BinaryExpression' && node.operator == ">") {
                expType = ExpressionType.GreaterThan;
            }

            if (node.type === 'BinaryExpression' && node.operator == ">=") {
                expType = ExpressionType.GreaterThanOrEqual;
            }
        });

        return expType;
    }
}
